from . import Resource


class Server(Resource):
    """
        A representation of a server

        Args:
            base (base): See :py:class:`nflex_connector_utils.resource.Resource` for common resource args.
            cpu_hz (int): Optional CPU Speed in Hz
            cpu_cores (int): Optional number of CPU cores
            ram_b (int): Optional RAM size in bytes
            volumes_b: Optional  total volume size in bytes
            state: CMP state. Must be one of: ``unknown``, ``pending``, ``running``, ``shutting``, ``terminated``, ``stopping``, ``stopped``,
            provider_state (str): Optional provider description of the state, can be any text.
            instance_type (str): Optional text value describing the instance type
            is_virtual (bool): Optional, set to ``True`` if the server is physical. Defaults to ``False``.
            image_detail (:py:class:`nflex_connector_utils.image_detail.ImageDetail`): An optional :py:class:`nflex_connector_utils.image_detail.ImageDetail` object
            ip_addresses (:py:class:`nflex_connector_utils.ip_address.IpAddress`): An optional :py:class:`nflex_connector_utils.ip_address.IpAddress` object


    """  # noqa

    def __init__(self, cpu_hz=None, cpu_cores=None, ram_b=None, volumes_b=None,
                 state=None, provider_state=None, image_detail=None,
                 instance_type=None, ip_addresses=None,
                 is_virtual=None, **kwargs):
        super(Server, self).__init__(type='server', **kwargs)
        self._cpu_cores = cpu_cores
        self._cpu_hz = cpu_hz
        self._ram_b = ram_b
        self._volumes_b = volumes_b
        self.is_virtual = is_virtual

        if state is None:
            state = 'unknown'
        self._state = state

        if provider_state is None:
            provider_state = 'unknown'
        self._provider_state = provider_state

        self._instance_type = instance_type
        self._image_detail = image_detail

        if ip_addresses is None:
            ip_addresses = []
        self._ip_addresses = ip_addresses

    def serialize(self):
        """Serialize the contents"""

        data = super(Server, self).serialize()

        ip_addresses = [ip.serialize() for ip in self._ip_addresses]

        data['details'] = {
            self.type: {
                "cpu_cores": self._cpu_cores,
                "cpu_hz": self._cpu_hz,
                "ram_b": self._ram_b,
                "volumes_b": self._volumes_b,
                "instance_type": self._instance_type,
                "state": self._state,
                "provider_state": self._provider_state,
                "ip_addresses": ip_addresses,
            }
        }

        type_details = data['details'][self.type]

        if self._image_detail is not None:
            type_details['image_detail'] = self._image_detail.serialize()
        else:
            type_details['image_detail'] = None

        if self.is_virtual is not None:
            type_details['is_virtual'] = self.is_virtual

        return data



from datetime import (
    datetime,
    timedelta,
)

_DEFAULT_TIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'
_DEFAULT_POLL_INTERVAL = 60 * 15  # 15 minutes


def setup_time_interval(event, backfill_time=None, initial_interval=0,
                        skew=0, time_format=_DEFAULT_TIME_FORMAT, logger=None):

    """
    sets up a time interval based on a combination of event data (last_update)
    and provider restriction (skew logic)

    Returns:
        (tuple):
            * start_time (datetime): interval start
            * end_time (datetime): interval end

    Args:

        event (dict): module's event that contains:

            * last_update (datetime): event's last_update which specifies the last time when data was fetched from provider
            * poll_interval (int): event's poll-interval in seconds which describes the time interval on when to fetch data from the provider (time of the execution - poll_interval)

        initial_interval (int): additional interval parameter in seconds to add to existing poll interval when generating interval start time (optional)

        backfill_time (int): the time required in seconds to backfill to ensure no data is missed (optional)

        time_format (str): the time format used to parse last_update datetime optional and defaults to %Y-%m-%dT%H:%M:%S.%fZ format

        skew (int): time in seconds used to discount from end time (optional)

        logger (option): nflex connector util's :class:`~nflex_connector_utils.logger` that allows to log info regarding usage of last_update and backfill_time


    Examples:
        set up time interval::

            event = {
                'last_update':datetime.utcnow(),
                'resource': {
                    'poll_interval': 15*60,
                },
            }

        (optional) set up logger::

            # get these values from nflex handler which has event and context
            logger = Logger(
                context=context,
                customer_id=customer_id
                account_id=account_id,
                resource_id=resource_id,
            )

        setup time interval::

            output = setup_time_interval(
                event=event,
                backfill_time=3600,
                initial_interval=5*60,
                skew=5*60,
                logger=logger,
            )

        parse and use the values::

            start_time, end_time = output

    """  # noqa

    last_update = event.get('last_update', None)
    poll_interval = event['resource']. \
        get('poll-interval', _DEFAULT_POLL_INTERVAL)

    info = ''
    end_time = datetime.utcnow() - timedelta(seconds=skew)

    if last_update is None:
        start_time = end_time - timedelta(
            seconds=(poll_interval + initial_interval)
        )
        info += "no last_update specified and use %d mins prior to " \
                "now as start_time\n" \
                % ((poll_interval + skew) / 60)

    else:
        start_time = datetime.strptime(last_update, time_format)
        diff = (end_time - start_time).total_seconds()
        if backfill_time and diff > backfill_time:
            end_time = start_time + timedelta(seconds=backfill_time)
            info += "Backfilling start_time %s end_time %s" \
                    "\nMax backfill time is %s hours" \
                    % (start_time, end_time, backfill_time / 3600)

        if logger:
            logger.info(info)

    return start_time, end_time



'''
Exceptions predefined here will be handled in spend collector.
'''


class InvalidCredentials(Exception):
    """
        Exception for invalid account credentials

        Args:
            msg (str): customised error msg.

        Returns:
            Raise an exception with either default message or given message
    """

    def __init__(self, msg=None):
        msg = msg or 'Invalid Credentials'
        super(InvalidCredentials, self).__init__(msg)


class InsufficientPermission(Exception):
    """
        Exception for insufficient permission. Raise this if the account does
        not have permission to access billing data

        Args:
            msg (str): customised error msg.

        Returns:
            Raise an exception with either default message or given message
    """
    def __init__(self, msg=None):
        msg = msg or 'Insufficient permission'
        super(InsufficientPermission, self).__init__(msg)


class NoBillingReport(Exception):
    """
        Exception for no billing report file found

        Args:
            msg (str): customised error msg.

        Returns:
            Raise an exception with either default message or given message
    """
    def __init__(self, msg=None):
        msg = msg or 'No billing report found'
        super(NoBillingReport, self).__init__(msg)


class SpendError(Exception):
    """
        Exception for any other kind of spend errors
    """
    pass



class Logger():
    """ Nflex Logger is simple wrapper for context.log which is
        supposed to be used in nflex module.
        This logger is achieves the following:

        - Each log message has header including customer_id, account_id, resource_id
          (optional).
        - The output destination can be changed easily between context.log and print
        - When you execute via flexer or GUI for test, you need to use "print" so you
          can see log from the result of flexer api(GUI used this api implicitly) you
          need to use "context.log" so you can see log at module status page. Through
          the context.module_id we dynamically know when it should use the logs
          API as this parameter is only set on production environments.

        Args:

            context (dict): nflex module's context
            customer_id (str): event's customer id
            account_id (str): event's account id
            resource_id (str): event's resource id (nullable)

        Examples:
            Set up logger::

                # Nflex Module Context
                context= {}

                # Nflex
                event = {
                    'customer_id': 'fake-id',
                    'account_id': 'fake-id',
                    'resource': {
                        'id': 'fake-resource-id'
                    }
                }

                logger = Logger(
                    context,
                    customer_id=event.get('customer_id', ''),
                    account_id=event.get('account_id', ''),
                    resource_id=event['resource']['id']
                )

            Log information with severity info, warn or error::

                logger.info('logging info %s' % ('something'))
                logger.warn('logging warn %s' % ('something'))
                logger.error('logging error %s' % ('something'))

    """  # noqa
    def __init__(self, context, customer_id, account_id, resource_id=None):
        self.context = context
        self.customer_id = customer_id
        self.account_id = account_id
        self.resource_id = resource_id
        self.template = "Target(customer: {customer_id},"\
            "account: {account_id}, resource: {resource_id}): {body}"

        severities = ['info', 'warn', 'error', 'exception']
        from functools import partial
        for severity in severities:
            setattr(self, severity, partial(self._log, severity))

    def _log(self, severity, body):
        prettied = self.template.format(
            customer_id=self.customer_id,
            account_id=self.account_id,
            resource_id=self.resource_id,
            body=body
            )
        if self.context.module_id:
            self.context.log(prettied, severity)
        else:
            print(severity + '\t' + prettied)



import httplib


class Account(object):
    """
        A representation of an Account
    """

    def __init__(self, id=None, metadata=None):
        self.id = id
        self.metadata = metadata

    def serialize(self):
        """Serialize the contents"""

        results = {}
        if self.id:
            results["id"] = self.id

        if self.metadata:
            results["metadata"] = self.metadata.serialize()

        return results

    def update(self, api, account_id):
        """Updates CMP account"""

        if account_id is None:
            '''For testing purposes, if the event doesn't have an account_id,
               don't bother trying to POST to the accounts API.'''
            return

        response = api.patch(
            path='/accounts/%s/resource' % account_id,
            data=self.serialize())
        if response.status_code != httplib.OK:
            raise Exception(
                'Got bad response from account API: %d, %s' % (
                    response.status_code, response.content))



from __future__ import division
from pyparsing import (  # noqa
    CaselessKeyword,
    CaselessLiteral,
    Forward,
    OneOrMore,
    Optional,
    ParseException,
    ParserElement,
    Regex,
    StringEnd,
    Word,
    alphanums,
    alphas,
    infixNotation,
    oneOf,
    opAssoc,
)

import logging
import operator
import os
import yaml

_CONVERSION_KEY = 'conversion_expr'

_DEFAULT_FILE_NAME = 'mapping.yaml'


def load_metric_mapping(file_path=_DEFAULT_FILE_NAME):
    """
    Loads a selection of metric definition mappings
    using SimpleExpressionParser.
    It does so by mapping values per metric key to ParsedEntry value which
    looks for conversion_expr in each metric definition and runs
    the logic on that field per evaluate mechanism.

    :return: mapping: mapping of metric with conversion expression ready to be evaluated.
    :rtype: dict of str: :class:`.ParsedEntry`

    Args:
        file_path (str): path to file (optional)

    Examples:
        read mappings from mappings.yaml from current directory (or absolute path)::

            metric1:
                name: metric1
                unit: "%"
                conversion_expr: value * 10

        Get original value::

            value = 10

        Load mapping::

            # can specify custom path by passing argument file_path
            # example file_path='my_project/definition.yaml'
            mapping = load_metric_mapping()

        Evaluate metric based on initial value:

            See :class:`.ParsedEntry`

    """  # noqa

    mapping = os.path.abspath(file_path)

    with open(mapping) as f:
        mapping = yaml.safe_load(f)

    parser = SimpleExpressionParser()

    parsed_mapping = dict()

    for key, val in mapping.iteritems():
        conversion = None
        if _CONVERSION_KEY in val:
            conversion = parser.parse(
                val[_CONVERSION_KEY]
            )

        parsed_mapping[key] = ParsedEntry(
            name=val.get('name', ''),
            unit=val.get('unit', ''),
            counter=val.get('counter', False),
            conversion=conversion,
        )

    return parsed_mapping


class ParsedEntry(object):
    """Represents a parsed mapping entry"""

    def __init__(self, name, unit, counter, conversion=None):
        self._name = name
        self._unit = unit
        self._counter = counter
        self._conversion = conversion

    def convert(self, **kwargs):
        """
        Different syntax for :func:`value`

        Evaluate metric based on initial value::

                metric = mapping['metric1']
                try:
                    # 10 * 10 = 100
                    value = metric.convert(
                        value=10,
                    )
                except VariableLookupError as e:
                    raise e

                name = metric.name()
                unit = metric.unit()
                counter = metric.counter()
        """
        return self.value(**kwargs)

    def value(self, **kwargs):
        """
        Evaluates conversion expression if it has one based on kwargs input.
        :raises VariableLookupError: failed to evaluate expression based on variables given.

        :return: value: evaluated value or original value if definition has no conversion expression

        Args:
            **kwargs(dict) - variables to evaluate on conversion_exr property of specified definition. The original value should be expressed as kwargs['value']

        Examples:

            See :func:`.load_metric_mapping` for load example

            Evaluate metric based on initial value::

                metric = mapping['metric1']
                try:
                    # 10 * 10 = 100
                    value = metric.value(
                        value=10,
                    )
                except VariableLookupError as e:
                    raise e

                name = metric.name()
                unit = metric.unit()
                counter = metric.counter()
        """  # noqa

        if not self._conversion:
            try:
                return kwargs['value']
            except KeyError as e:
                raise VariableLookupError(e)

        return self._conversion.evaluate(kwargs)

    def unit(self):
        """Retrieves metric's unit"""
        return self._unit

    def name(self):
        """Retrieves metric's name"""
        return self._name

    def counter(self):
        """Retrieves metric's counter"""
        return self._counter


class ExpressionParser(object):
    """Takes care of parsing simple math expressions with variables"""
    def __init__(self):
        # a class-level static method to enable a memoizing performance
        # enhancement, known as "packrat parsing".
        ParserElement.enablePackrat()

        NUMBER = Regex(r"[+-]?\d+(:?\.\d*)?(:?[eE][+-]?\d+)?").setParseAction(
            Immediate
        )
        IDENT = Word(alphas, alphanums + '_').setParseAction(Variable)
        self.OPERAND = (NUMBER | IDENT)
        self.MATH_OPERATORS = [
            (oneOf('+ -'), 1, opAssoc.RIGHT, SignTerm),
            ('^', 2, opAssoc.RIGHT, MathTerm),
            (oneOf('* /'), 2, opAssoc.LEFT, MathTerm),
            (oneOf('+ -'), 2, opAssoc.LEFT, MathTerm)
        ]
        self.ARITH_EXPR = infixNotation(self.OPERAND, self.MATH_OPERATORS)
        self.pattern = self.ARITH_EXPR + StringEnd()

    def parse(self, expression):
        try:
            return self.pattern.parseString(expression)[0]

        except ParseException as e:
            logging.warn('Failed to parse "%s": %s' % (expression, e))

        return None


class TernaryExpressionParser(ExpressionParser):
    """Takes care of parsing Python ternary statements with variables"""
    def __init__(self):
        super(TernaryExpressionParser, self).__init__()
        IF = CaselessKeyword('if').suppress()
        ELSE = CaselessKeyword('else').suppress()
        RELATIONAL_OPERATOR = oneOf(">= <= != > < = ==")
        CONDITION = (self.ARITH_EXPR +
                     RELATIONAL_OPERATOR +
                     self.ARITH_EXPR).setParseAction(Condition)
        self.TERNARY_EXPR = (
            self.ARITH_EXPR.setResultsName('if_expr') +
            IF +
            CONDITION.setResultsName('condition') +
            ELSE +
            self.ARITH_EXPR.setResultsName('else_expr')
        ).setParseAction(TernaryExpression)
        self.pattern = self.TERNARY_EXPR + StringEnd()


class CaseExpressionParser(TernaryExpressionParser):
    """Takes care of parsing (nested) SQL CASE statements with variables"""
    def __init__(self):
        super(CaseExpressionParser, self).__init__()
        CASE = CaselessKeyword('case').suppress()
        WHEN = CaselessKeyword('when').suppress()
        THEN = CaselessKeyword('then').suppress()
        ELSE = CaselessKeyword('else').suppress()
        END = CaselessKeyword('end').suppress()
        NOT = CaselessLiteral('not')
        AND = CaselessLiteral('and')
        OR = CaselessLiteral('or')

        RELATIONAL_OPERATOR = oneOf(">= <= != > < =")

        self.CASE_EXPR = Forward()
        # Case statements can be operands
        self.OPERAND = self.CASE_EXPR | self.OPERAND
        ARITH_EXPR = infixNotation(self.OPERAND, self.MATH_OPERATORS)

        CONDITION = (ARITH_EXPR +
                     RELATIONAL_OPERATOR +
                     ARITH_EXPR).setParseAction(Condition)

        WHEN_TERM = infixNotation(CONDITION,
                                  [(NOT, 1, opAssoc.RIGHT, NegativeTerm),
                                   (AND, 2, opAssoc.LEFT, AndTerm),
                                   (OR, 2, opAssoc.LEFT, OrTerm)])
        WHEN_EXPR = WHEN + WHEN_TERM.setResultsName('when_expr',
                                                    listAllMatches=True)

        THEN_EXPR = THEN + ARITH_EXPR.setResultsName('then_expr',
                                                     listAllMatches=True)

        ELSE_EXPR = ELSE + ARITH_EXPR.setResultsName('else_expr')

        self.CASE_EXPR << (
            CASE +
            OneOrMore(WHEN_EXPR + THEN_EXPR) + Optional(ELSE_EXPR) +
            END
        ).setParseAction(CaseStatement)
        self.pattern = self.CASE_EXPR + StringEnd()


class SimpleExpressionParser(CaseExpressionParser):
    """Takes care of parsing only python ternary
    expressions and simple math expressions with variables
    """
    def __init__(self):
        super(self.__class__, self).__init__()
        self.pattern = (self.TERNARY_EXPR |
                        self.ARITH_EXPR) + StringEnd()


class Immediate(object):
    def __init__(self, tokens):
        self.value = tokens[0]

    def evaluate(self, variables):
        return float(self.value)

    def __repr__(self):
        return '<Immediate value=%s>' % self.value


class Variable(object):
    def __init__(self, tokens):
        self.value = tokens[0]

    def evaluate(self, variables):
        if self.value == 'NULL':
            return None
        try:
            return float(variables[self.value])

        except KeyError:
            raise VariableLookupError('Lookup failed for variable %s in %s'
                                      % (self.value, variables))

    def __repr__(self):
        return '<Variable value=%s>' % self.value


class Condition(object):
    opMap = {
        '<': operator.lt,
        '<=': operator.le,
        '>': operator.gt,
        '>=': operator.ge,
        '=': operator.eq,
        '==': operator.eq,
        '!=': operator.ne,
    }

    def __init__(self, tokens):
        self.left, self.op, self.right = tokens

    def evaluate(self, variables):
        return self.opMap[self.op](self.left.evaluate(variables),
                                   self.right.evaluate(variables))


class CaseStatement(object):
    def __init__(self, tokens):
        self.if_then_pairs = zip(tokens['when_expr'], tokens['then_expr'])
        self.else_value = tokens.get('else_expr')

    def evaluate(self, variables):
        result = 'NULL'
        # go through all WHEN-THEN pairs and try to find a match
        for pair in self.if_then_pairs:
            if pair[0].evaluate(variables):
                result = pair[1].evaluate(variables)
                break

        # if none of the WHEN clauses are matched, use the ELSE clause
        if result == 'NULL':
            result = None
            if self.else_value:
                result = self.else_value.evaluate(variables)

        return result


class TernaryExpression(object):
    def __init__(self, tokens):
        self.condition = tokens['condition']
        self.if_expr = tokens['if_expr']
        self.else_expr = tokens['else_expr']

    def evaluate(self, variables):
        result = None
        if self.condition.evaluate(variables):
            result = self.if_expr.evaluate(variables)

        else:
            result = self.else_expr.evaluate(variables)

        return result


class BoolTerm(object):
    def __init__(self, tokens):
        self.operands = tokens[0][0::2]

    def evaluate(self, variables):
        return self.operator(o.evaluate(variables) for o in self.operands)


class AndTerm(BoolTerm):
    operator = all


class OrTerm(BoolTerm):
    operator = any


class NegativeTerm(object):
    def __init__(self, tokens):
        self.operand = tokens[0][1]

    def evaluate(self, variables):
        return not self.operand.evaluate(variables)


class MathTerm(object):
    opMap = {
        '+': operator.add,
        '-': operator.sub,
        '*': operator.mul,
        '/': operator.truediv,
        '^': operator.pow,
    }

    def __init__(self, tokens):
        self.operands = tokens[0][0::2]
        self.operators = tokens[0][1::2]

    def evaluate(self, variables):
        result = self.operands[0].evaluate(variables)
        for i, op in enumerate(self.operators, start=1):
            try:
                result = self.opMap[op](result,
                                        self.operands[i].evaluate(variables))

            except ZeroDivisionError:
                result = 0.0

        return result


class SignTerm(object):
    def __init__(self, tokens):
        self.sign, self.operand = tokens[0]

    def evaluate(self, variables=None):
        if self.sign == '+':
            return self.operand.evaluate(variables)
        else:
            return -self.operand.evaluate(variables)


class VariableLookupError(Exception):
    """ Occurs when a variable lookup fails """
